# -*- coding: utf-8 -*-

from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

import numpy as np
from utils import debug

from tocpushbutton import TocPushButton

class TocManager(QtCore.QObject):
    tocIndexChanged = QtCore.pyqtSignal(int)
    def __init__(self, tocButton):
        parent = tocButton.parentWidget()
        super(TocManager, self).__init__(tocButton)

        self.view = QtWidgets.QTreeView(parent)
        self.model = QtGui.QStandardItemModel(parent)
        self.view.setModel(self.model)
        self.view.clicked.connect(self.OnViewClicked)
        self.view.hide()

        self.tocButton = tocButton
        self.tocButton.setView(self.view)

        self.toc = None

    def setToc(self, toc):
        # remove all model items first
        self.model.removeRows(0, self.model.rowCount())

        # clear title
        self.tocButton.clearTitleText()

        self.toc = toc
        self.toc_item_parents = None
        self.toc_item_pages = None
        self.toc_model_items = None
        self.current_page_idx = -1

        # extract parent informaton for each node from toc (in format of PyMuPDF)
        tocItemCnt = len(toc)
        self.toc_item_parents = [-1] * tocItemCnt # each item has one parent
        self.toc_item_pages = []
        parent_queue = []
        for i in range(tocItemCnt):
            lvl, title, page, extra = toc[i]
            self.toc_item_pages.append(page - 1) # make page start from 0
            if i == 0:
                assert(lvl == 1)
                parent_queue.append(-1)
                continue
            # 
            lvl_prev = toc[i-1][0]
            if lvl > lvl_prev:
                assert(lvl == lvl_prev + 1)
                parent_queue.append(i-1)
            else:
                for j in range(lvl_prev - lvl):
                    parent_queue.pop(-1)
            self.toc_item_parents[i] = parent_queue[-1] # the latest one
        self.toc_item_parents = np.array(self.toc_item_parents)
        self.toc_item_pages = np.array(self.toc_item_pages)

        # construct tree structure
        rooItem = self.model.invisibleRootItem()
        self.toc_model_items = [None] * tocItemCnt
        for i in range(tocItemCnt):
            lvl, title, page, extra = toc[i] # page start from 1
            item = QtGui.QStandardItem(title)
            item.setData(page - 1, QtCore.Qt.UserRole + 1)
            # 
            pageItem = item.clone()
            if page >= 1:
                pageItem.setText(str(page))
            else:
                pageItem.setText("")
            pageItem.setTextAlignment(QtCore.Qt.AlignRight)
            pid = self.toc_item_parents[i]
            if pid == -1:
                rooItem.appendRow([item, pageItem])
            else:
                self.toc_model_items[pid].appendRow([item, pageItem])
            self.toc_model_items[i] = item

        # 
        if tocItemCnt > 0:
            self.view.setAlternatingRowColors(True)
            self.view.setHeaderHidden(True)
            self.view.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
            self.view.header().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
            self.view.header().setStretchLastSection(False)

    def OnViewClicked(self, mIndex):
        page_no = self.view.model().data(mIndex, QtCore.Qt.UserRole+1)
        if page_no >= 0:
            self.tocIndexChanged.emit(page_no)
            self.view.hide()

    def update(self, current_page_idx):
        if not self.toc:
            return

        if current_page_idx == self.current_page_idx:
            return

        self.current_page_idx = current_page_idx

        # get the first one larger than current_page_idx
        current_toc_idx = np.argmax(self.toc_item_pages > self.current_page_idx)
        if self.toc_item_pages[current_toc_idx] > self.current_page_idx: # found successfully
            if current_toc_idx > 0:
                current_toc_idx -= 1 # get the previous one
        else: # can not find
            current_toc_idx = len(self.toc) - 1 # get the very last one
            
        # 
        scrollTargetModelIndex = self.model.indexFromItem(self.toc_model_items[current_toc_idx])

        # 
        title_list = []
        valid_tocs = []
        while current_toc_idx >= 0:
            lvl, title, page, extra = self.toc[current_toc_idx]
            title_list.append(title)
            valid_tocs.append(current_toc_idx)
            current_toc_idx = self.toc_item_parents[current_toc_idx]
        title_list.reverse()
        self.tocButton.setTitleText(title_list)

        self.view.clearSelection()
        self.view.expand(scrollTargetModelIndex)
        self.view.scrollTo(scrollTargetModelIndex, QtWidgets.QAbstractItemView.ScrollHint.PositionAtCenter)
        self.view.selectionModel().setCurrentIndex(scrollTargetModelIndex, QtCore.QItemSelectionModel.Select)



from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

import numpy as np
from utils import debug

class viewEventFilter(QtCore.QObject):
    def eventFilter(self, widget, event):
        # FocusOut event
        if event.type() == QtCore.QEvent.FocusOut:
            # do custom stuff
            # print('focus out')
            widget.hide()
            # return False so that the widget will also handle the event
            # otherwise it won't focus out
            return False
        else:
            # we don't care about other events
            return False

class TocPushButton(QtWidgets.QPushButton):
    def __init__(self, parent):
        super(TocPushButton, self).__init__(parent)
        self.parent = parent
        self.clicked.connect(self.onClicked)
        self.view = None
        self.title_list = []

    def setView(self, view):
        self.view = view
        self.view_event_filter = viewEventFilter()
        self.view.installEventFilter(self.view_event_filter)

    def resizeEvent(self, ev):
        self.updateTitleText()
        # 
        rect = self.geometry()
        self.view.setGeometry(QtCore.QRect(0, rect.height(), self.parent.width(), 400))
        super(TocPushButton, self).resizeEvent(ev)

    def onClicked(self):
        # debug("onClicked in TocButton")
        if self.view:
            if self.view.isVisible():
                self.view.hide()
            else:
                self.view.show()
                self.view.setFocus()

    def constructTitleText(self, title_list, cutLength=10000):
        title_str = ""
        for title in title_list:
            title_str += " ⯈ "
            tmpStr = title[:cutLength]
            if len(title) > cutLength:
                tmpStr += " ..."
            title_str += tmpStr
        title_str += " "
        return title_str

    def updateTitleText(self):
        if len(self.title_list) == 0:
            return
        # https://stackoverflow.com/questions/8633433/qt-how-to-get-the-pixel-length-of-a-string-in-a-qlabel
        titleStr = self.constructTitleText(self.title_list)
        textWidth = self.fontMetrics().boundingRect(titleStr).width()
        if textWidth > self.width() - 10:
            titleStr = self.constructTitleText(self.title_list, cutLength=12)
        # 
        textWidth = self.fontMetrics().boundingRect(titleStr).width()
        if textWidth > self.width() - 10:
            self.setStyleSheet("text-align:right")
        else:
            self.setStyleSheet("text-align:left")
        self.setText(titleStr)

    def setTitleText(self, title_list):
        self.title_list = title_list
        self.updateTitleText()

    def clearTitleText(self):
        self.title_list = []
        self.setText(" ⯈ ")
        self.setStyleSheet("text-align:left")


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys, os

from subprocess import Popen

from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

# add parent directory to path
currentDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.dirname(currentDir)
# debug(currentDir)
# debug(parentDir)
sys.path.append(currentDir)
# sys.path.append(parentDir)
sys.path.append(currentDir + os.path.sep + 'resources')

from utils import debug

from resources.ui_main import Ui_window

from document import DocumentView
from library import LibraryView

from dialogs import ExportToImageDialog, DocInfoDialog

from docgraphicsview import DocGraphicsView

HOMEDIR = os.path.expanduser("~")

class MainWindow(QtWidgets.QMainWindow, Ui_window):
    def __init__(self, screens, app_data_path):
        super(MainWindow, self).__init__() # Call the inherited classes __init__ method
        self.setupUi(self)

        self.screen_dpi = screens[0].logicalDotsPerInch()
        self.app_data_path = app_data_path
        
        self.libWidget = LibraryView(self.centraltabwidget, self.screen_dpi, self.app_data_path)
        self.libWidget.fileReselected.connect(self.onFileReselected)
        self.libWidget.showStatusRequested.connect(self.onShowStatusRequested)
        self.centraltabwidget.addTab(self.libWidget, "My Library")
        
        self.centraltabwidget.setTabsClosable(True)

        # https://stackoverflow.com/questions/2616483/close-button-only-for-some-tabs-in-qt
        # make the first tab unclosable
        self.centraltabwidget.tabBar().tabButton(0, QtWidgets.QTabBar.RightSide).resize(0,0)

        self.centraltabwidget.currentChanged.connect(self.onTabChanged)
        self.centraltabwidget.tabCloseRequested.connect(self.onTabClose)

        self.centraltabwidget.setStyleSheet("QTabWidget::pane {margin: 0 0 0 0}")

        # self.dockSearch.hide()
        # self.dockWidget.hide()
        # self.dockWidget.setMinimumWidth(310)
        # self.findTextEdit.setFocusPolicy(QtCore.Qt.StrongFocus)

        # resizing pages requires some time to take effect
        # self.resize_page_timer = QtCore.QTimer(self)
        # self.resize_page_timer.setSingleShot(True)
        # self.resize_page_timer.timeout.connect(self.onWindowResize)
        # Add shortcut actions
        # self.gotoPageAction = QtWidgets.QAction(QIcon(":/goto.png"), "GoTo Page", self)
        # self.gotoPageAction.triggered.connect(self.gotoPage)
        # self.copyTextAction = QtWidgets.QAction(QIcon(":/copy.png"), "Copy Text", self)
        # self.copyTextAction.setCheckable(True)
        # self.copyTextAction.triggered.connect(self.toggleCopyText)
        # self.findTextAction = QtWidgets.QAction(QIcon(":/search.png"), "Find Text", self)
        # self.findTextAction.setShortcut('Ctrl+F')
        # self.findTextAction.triggered.connect(self.dockSearch.show)

        # connect menu actions signals
        self.openFileAction.triggered.connect(self.openFile)
        self.libWidget.pushButton_open.clicked.connect(self.openFile)
        
        # self.pageToImageAction.triggered.connect(self.exportPageToImage)
        # self.docInfoAction.triggered.connect(self.docInfo)
        # self.toPSAction.triggered.connect(self.exportToPS)
        # self.printAction.triggered.connect(self.printFile)

        # # Add widgets
        # self.statusbar = QLabel(self)
        # self.statusbar.setStyleSheet("QLabel { font-size: 12px; border-radius: 2px; padding: 2px; background: palette(highlight); color: palette(highlighted-text); }")
        # self.statusbar.setMaximumHeight(16)
        # self.statusbar.hide()

        # Import settings
        desktop = QtWidgets.QApplication.desktop()
        self.settings = QtCore.QSettings("kuafu", "main", self)
        self.recent_files = self.settings.value("RecentFiles", [])
        self.history_filenames = self.settings.value("HistoryFileNameList", [])
        self.history_page_no = self.settings.value("HistoryPageNoList", [])
        geometry = self.settings.value('geometry', bytes('', 'utf-8'))
        self.restoreGeometry(geometry)
        
        state = self.settings.value('library/spliter', bytes('', 'utf-8'))
        self.libWidget.splitter.restoreState(state)
        # state = self.settings.value('library/spliter_doc', bytes('', 'utf-8'))
        # self.libWidget.splitter_doc.restoreState(state)

        # self.offset_x = int(self.settings.value("OffsetX", 4))
        # self.offset_y = int(self.settings.value("OffsetY", 26))
        # self.available_area = [desktop.availableGeometry().width(), desktop.availableGeometry().height()]
        # self.zoomLevelCombo.setCurrentIndex(int(self.settings.value("ZoomLevel", 2)))

        # Connect Signals
        # self.findTextEdit.returnPressed.connect(self.findNext)
        # self.findNextButton.clicked.connect(self.findNext)
        # self.findBackButton.clicked.connect(self.findBack)
        # self.findCloseButton.clicked.connect(self.dockSearch.hide)
        # self.dockSearch.visibilityChanged.connect(self.toggleFindMode)

        self.recent_files_actions = []
        self.addRecentFiles()

        self.show()

        self.loadPDFfile(self.recent_files[0]) # load the latest one

    def onShowStatusRequested(self, msg):
        self.showStatus(msg)

    def showStatus(self, msg):
        self.statusBar.showMessage(msg, msecs=5000)
        # if url=="":
        #     self.statusbar.hide()
        #     return
        # self.statusbar.setText(url)
        # self.statusbar.adjustSize()
        # self.statusbar.move(0, self.height()-self.statusbar.height())
        # self.statusbar.show()

    def onActionFind(self):
        self.findTxtEdit.setFocus()

    def onFileReselected(self, filefullpath):
        filename = os.path.split(filefullpath)[1]
        self.setWindowTitle(filename + ' -- kuafu')

    def addRecentFiles(self):
        self.recent_files_actions[:] = [] # pythonic way to clear list
        self.menuRecentFiles.clear()
        for each in self.recent_files:
            name = elideMiddle(os.path.basename(each), 60)
            action = self.menuRecentFiles.addAction(name, self.openRecentFile)
            self.recent_files_actions.append(action)
        self.menuRecentFiles.addSeparator()
        self.menuRecentFiles.addAction(QtGui.QIcon(':/edit-clear.png'), 'Clear Recents', self.clearRecents)

    def openRecentFile(self):
        action = self.sender()
        index = self.recent_files_actions.index(action)
        self.loadPDFfile(self.recent_files[index])

    def clearRecents(self):
        self.recent_files_actions[:] = []
        self.menuRecentFiles.clear()
        self.recent_files[:] = []

    def removeOldDoc(self):
        # Save current page number
        self.saveFileData()
        # Remove old document
        for i in range(len(self.pages)):
            self.verticalLayout.removeWidget(self.pages[-1])
        for i in range(len(self.pages)):
            self.pages.pop().deleteLater()
        self.frame.deleteLater()
        self.jumped_from = None
        #self.addRecentFiles()

    def loadPDFfile(self, filename):
        """ Loads pdf document in all threads """
        filename = os.path.expanduser(filename)
        
        # widget = DocumentView(self.centraltabwidget, filename, self.screen_dpi)
        # widget.showStatusRequested.connect(self.showStatus)
        # self.centraltabwidget.addTab(widget, filename)
        # self.centraltabwidget.setCurrentWidget(widget)

        # self.annotview_model.removeRows(0, self.annotview_model.rowCount())
        self.libWidget.setDocument(filename, self.screen_dpi)
        

    def openFile(self):
        filename, sel_filter = QtWidgets.QFileDialog.getOpenFileName(self,
                                      "Select Document to Open", "",
                                      "Portable Document Format (*.pdf);;All Files (*)" )
        if filename != "":
            self.loadPDFfile(filename)

    def printFile(self):
        Popen(["quikprint", self.filename])

    def exportToPS(self):
        width = self.doc.page(self.current_page-1).pageSizeF().width()
        height = self.doc.page(self.current_page-1).pageSizeF().height()
        filename, sel_filter = QtWidgets.QFileDialog.getSaveFileName(self, "Select File to Save",
                                       os.path.splitext(self.filename)[0]+'.ps',
                                      "Adobe Postscript Format (*.ps)" )
        if filename == '' : return
        conv = self.doc.psConverter()
        conv.setPaperWidth(width)
        conv.setPaperHeight(height)
        conv.setOutputFileName(filename)
        conv.setPageList([i+1 for i in range(self.pages_count)])
        ok = conv.convert()
        if ok:
            QMessageBox.information(self, "Successful !","File has been successfully exported")
        else:
            QMessageBox.warning(self, "Failed !","Failed to export to Postscript")

    def exportPageToImage(self):
        dialog = ExportToImageDialog(self.current_page, self.pages_count, self)
        if dialog.exec_() == QDialog.Accepted:
            try:
                dpi = int(dialog.dpiEdit.text())
                page_no = dialog.pageNoSpin.value()
                filename = os.path.splitext(self.filename)[0]+'-'+str(page_no)+'.jpg'
                page = self.doc.page(page_no-1)
                if not page : return
                img = page.renderToImage(dpi, dpi)
                img.save(filename)
                QMessageBox.information(self, "Successful !","Page has been successfully exported")
            except:
                QMessageBox.warning(self, "Failed !","Failed to export to Image")

    def onTabChanged(self, tabIdx):
        debug("tab changed")
        widget = self.centraltabwidget.widget(tabIdx)
        if isinstance(widget, DocumentView):
            currentPage = widget.current_page
            currentPageCount = widget.pages_count

    def saveFileData(self, filename, current_page):
        if filename != '':
            filename = collapseUser(filename)
            if filename in self.history_filenames:
                index = self.history_filenames.index(filename)
                self.history_page_no[index] = current_page
            else:
                self.history_filenames.insert(0, filename)
                self.history_page_no.insert(0, current_page)
            if filename in self.recent_files:
                self.recent_files.remove(filename)
            self.recent_files.insert(0, filename)

    def onTabClose(self, tabIdx):
        widget = self.centraltabwidget.widget(tabIdx)
        # save information if necessary
        # if isinstance(widget, DocumentView):
        # remove
        widget.close()
        self.centraltabwidget.removeTab(tabIdx)

    def onAppQuit(self):
        debug("OnAppQuit")
        """ Close running threads """
        tabCnt = self.centraltabwidget.count()
        for i in range(tabCnt):
            self.centraltabwidget.widget(i).close()

        # filename = widget.filename
        # current_page = widget.current_page
        if len(self.libWidget.filename) > 0:
            self.saveFileData(self.libWidget.filename, 0)
        # 
        geometry = self.saveGeometry()
        self.settings.setValue('geometry', geometry)

        state = self.libWidget.splitter.saveState()
        self.settings.setValue('library/spliter', state)
        # state = self.libWidget.splitter_doc.saveState()
        # self.settings.setValue('library/spliter_doc', state)

        # self.settings.setValue("OffsetX", self.geometry().x()-self.x())
        # self.settings.setValue("OffsetY", self.geometry().y()-self.y())
        # self.settings.setValue("ZoomLevel", self.zoom_level)
        if len(self.history_filenames) > 0:
            self.settings.setValue("HistoryFileNameList", self.history_filenames[:100])
        if len(self.history_page_no) > 0:
            self.settings.setValue("HistoryPageNoList", self.history_page_no[:100])
        if len(self.recent_files) > 0:
            self.settings.setValue("RecentFiles", self.recent_files[:10]) 

def wait(millisec):
    loop = QtCore.QEventLoop()
    QtCore.QTimer.singleShot(millisec, loop.quit)
    loop.exec_()

def elideMiddle(text, length):
    if len(text) <= length: return text
    return text[:length//2] + '...' + text[len(text)-length+length//2:]

def collapseUser(path):
    ''' converts /home/user/file.ext to ~/file.ext '''
    if path.startswith(HOMEDIR):
        return path.replace(HOMEDIR, '~', 1)
    return path

def main():
    app = QtWidgets.QApplication(sys.argv)
    # 
    # https://stackoverflow.com/questions/32525196/how-to-get-a-settings-storage-path-in-a-cross-platform-way-in-qt
    app.setOrganizationDomain("huyinlin@gmail.com")
    app.setApplicationName("kuafu")
    app_data_path = QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.AppDataLocation)
    if not os.path.exists(app_data_path):
        os.makedirs(app_data_path)
    # 
    win = MainWindow(app.screens(), app_data_path)
    if len(sys.argv)>1 and os.path.exists(os.path.abspath(sys.argv[-1])):
        win.loadPDFfile(os.path.abspath(sys.argv[-1]))
    app.aboutToQuit.connect(win.onAppQuit)
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './kuafu/resources/ui_library.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_librarywidget(object):
    def setupUi(self, librarywidget):
        librarywidget.setObjectName("librarywidget")
        librarywidget.resize(1115, 701)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(librarywidget)
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_4.setSpacing(0)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.splitter = QtWidgets.QSplitter(librarywidget)
        self.splitter.setOrientation(QtCore.Qt.Horizontal)
        self.splitter.setObjectName("splitter")
        self.widget = QtWidgets.QWidget(self.splitter)
        self.widget.setMaximumSize(QtCore.QSize(600, 16777215))
        self.widget.setObjectName("widget")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.widget)
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setSpacing(0)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.pushButton_open = QtWidgets.QPushButton(self.widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_open.sizePolicy().hasHeightForWidth())
        self.pushButton_open.setSizePolicy(sizePolicy)
        self.pushButton_open.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/document-open.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_open.setIcon(icon)
        self.pushButton_open.setFlat(True)
        self.pushButton_open.setObjectName("pushButton_open")
        self.horizontalLayout_2.addWidget(self.pushButton_open)
        self.pushButton_save_all = QtWidgets.QPushButton(self.widget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_save_all.sizePolicy().hasHeightForWidth())
        self.pushButton_save_all.setSizePolicy(sizePolicy)
        self.pushButton_save_all.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/document-save.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_save_all.setIcon(icon1)
        self.pushButton_save_all.setFlat(True)
        self.pushButton_save_all.setObjectName("pushButton_save_all")
        self.horizontalLayout_2.addWidget(self.pushButton_save_all)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.fileview = QtWidgets.QTreeView(self.widget)
        self.fileview.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.fileview.setObjectName("fileview")
        self.verticalLayout.addWidget(self.fileview)
        self.widget1 = QtWidgets.QWidget(self.splitter)
        self.widget1.setObjectName("widget1")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.widget1)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSpacing(0)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.pushButton_prev = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_prev.sizePolicy().hasHeightForWidth())
        self.pushButton_prev.setSizePolicy(sizePolicy)
        self.pushButton_prev.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/go-previous.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_prev.setIcon(icon2)
        self.pushButton_prev.setFlat(True)
        self.pushButton_prev.setObjectName("pushButton_prev")
        self.horizontalLayout.addWidget(self.pushButton_prev)
        self.pushButton_next = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_next.sizePolicy().hasHeightForWidth())
        self.pushButton_next.setSizePolicy(sizePolicy)
        self.pushButton_next.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/go-next.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_next.setIcon(icon3)
        self.pushButton_next.setFlat(True)
        self.pushButton_next.setObjectName("pushButton_next")
        self.horizontalLayout.addWidget(self.pushButton_next)
        self.line = QtWidgets.QFrame(self.widget1)
        self.line.setMaximumSize(QtCore.QSize(16777215, 28))
        self.line.setFrameShape(QtWidgets.QFrame.VLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.horizontalLayout.addWidget(self.line)
        self.pushButton_zoomIn = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_zoomIn.sizePolicy().hasHeightForWidth())
        self.pushButton_zoomIn.setSizePolicy(sizePolicy)
        self.pushButton_zoomIn.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/zoom-in.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_zoomIn.setIcon(icon4)
        self.pushButton_zoomIn.setFlat(True)
        self.pushButton_zoomIn.setObjectName("pushButton_zoomIn")
        self.horizontalLayout.addWidget(self.pushButton_zoomIn)
        self.pushButton_zoomOut = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_zoomOut.sizePolicy().hasHeightForWidth())
        self.pushButton_zoomOut.setSizePolicy(sizePolicy)
        self.pushButton_zoomOut.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/icons/zoom-out.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_zoomOut.setIcon(icon5)
        self.pushButton_zoomOut.setFlat(True)
        self.pushButton_zoomOut.setObjectName("pushButton_zoomOut")
        self.horizontalLayout.addWidget(self.pushButton_zoomOut)
        self.pushButton_zoomFitWidth = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_zoomFitWidth.sizePolicy().hasHeightForWidth())
        self.pushButton_zoomFitWidth.setSizePolicy(sizePolicy)
        self.pushButton_zoomFitWidth.setText("")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/icons/zoom-fit-width.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_zoomFitWidth.setIcon(icon6)
        self.pushButton_zoomFitWidth.setCheckable(True)
        self.pushButton_zoomFitWidth.setFlat(True)
        self.pushButton_zoomFitWidth.setObjectName("pushButton_zoomFitWidth")
        self.horizontalLayout.addWidget(self.pushButton_zoomFitWidth)
        self.line_2 = QtWidgets.QFrame(self.widget1)
        self.line_2.setMaximumSize(QtCore.QSize(16777215, 28))
        self.line_2.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.horizontalLayout.addWidget(self.line_2)
        self.pushButton_oneColumn = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_oneColumn.sizePolicy().hasHeightForWidth())
        self.pushButton_oneColumn.setSizePolicy(sizePolicy)
        self.pushButton_oneColumn.setText("")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/icons/snap-page.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_oneColumn.setIcon(icon7)
        self.pushButton_oneColumn.setCheckable(True)
        self.pushButton_oneColumn.setFlat(True)
        self.pushButton_oneColumn.setObjectName("pushButton_oneColumn")
        self.horizontalLayout.addWidget(self.pushButton_oneColumn)
        self.pushButton_twoColumn = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_twoColumn.sizePolicy().hasHeightForWidth())
        self.pushButton_twoColumn.setSizePolicy(sizePolicy)
        self.pushButton_twoColumn.setText("")
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(":/icons/view-pages-facing.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_twoColumn.setIcon(icon8)
        self.pushButton_twoColumn.setCheckable(True)
        self.pushButton_twoColumn.setFlat(True)
        self.pushButton_twoColumn.setObjectName("pushButton_twoColumn")
        self.horizontalLayout.addWidget(self.pushButton_twoColumn)
        self.pushButton_fourColumn = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_fourColumn.sizePolicy().hasHeightForWidth())
        self.pushButton_fourColumn.setSizePolicy(sizePolicy)
        self.pushButton_fourColumn.setText("")
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap(":/icons/view-pages-overview.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_fourColumn.setIcon(icon9)
        self.pushButton_fourColumn.setCheckable(True)
        self.pushButton_fourColumn.setFlat(True)
        self.pushButton_fourColumn.setObjectName("pushButton_fourColumn")
        self.horizontalLayout.addWidget(self.pushButton_fourColumn)
        self.pushButton_emptyPage = QtWidgets.QPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pushButton_emptyPage.sizePolicy().hasHeightForWidth())
        self.pushButton_emptyPage.setSizePolicy(sizePolicy)
        self.pushButton_emptyPage.setText("")
        icon10 = QtGui.QIcon()
        icon10.addPixmap(QtGui.QPixmap(":/icons/page-simple.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pushButton_emptyPage.setIcon(icon10)
        self.pushButton_emptyPage.setCheckable(True)
        self.pushButton_emptyPage.setFlat(True)
        self.pushButton_emptyPage.setObjectName("pushButton_emptyPage")
        self.horizontalLayout.addWidget(self.pushButton_emptyPage)
        self.line_3 = QtWidgets.QFrame(self.widget1)
        self.line_3.setMaximumSize(QtCore.QSize(16777215, 28))
        self.line_3.setFrameShape(QtWidgets.QFrame.VLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.horizontalLayout.addWidget(self.line_3)
        self.tocButton_1 = TocPushButton(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tocButton_1.sizePolicy().hasHeightForWidth())
        self.tocButton_1.setSizePolicy(sizePolicy)
        self.tocButton_1.setMinimumSize(QtCore.QSize(0, 0))
        self.tocButton_1.setMaximumSize(QtCore.QSize(16777215, 32))
        self.tocButton_1.setFocusPolicy(QtCore.Qt.NoFocus)
        self.tocButton_1.setStyleSheet("text-align:left")
        self.tocButton_1.setFlat(True)
        self.tocButton_1.setObjectName("tocButton_1")
        self.horizontalLayout.addWidget(self.tocButton_1)
        self.lineEdit_pageNo = PageNoLineEdit(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.lineEdit_pageNo.sizePolicy().hasHeightForWidth())
        self.lineEdit_pageNo.setSizePolicy(sizePolicy)
        self.lineEdit_pageNo.setMaximumSize(QtCore.QSize(52, 32))
        self.lineEdit_pageNo.setStyleSheet("")
        self.lineEdit_pageNo.setObjectName("lineEdit_pageNo")
        self.horizontalLayout.addWidget(self.lineEdit_pageNo)
        self.label_of = QtWidgets.QLabel(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_of.sizePolicy().hasHeightForWidth())
        self.label_of.setSizePolicy(sizePolicy)
        self.label_of.setObjectName("label_of")
        self.horizontalLayout.addWidget(self.label_of)
        self.label_pageCount = QtWidgets.QLabel(self.widget1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_pageCount.sizePolicy().hasHeightForWidth())
        self.label_pageCount.setSizePolicy(sizePolicy)
        self.label_pageCount.setMinimumSize(QtCore.QSize(48, 0))
        self.label_pageCount.setMaximumSize(QtCore.QSize(48, 32))
        self.label_pageCount.setStyleSheet("")
        self.label_pageCount.setAlignment(QtCore.Qt.AlignCenter)
        self.label_pageCount.setObjectName("label_pageCount")
        self.horizontalLayout.addWidget(self.label_pageCount)
        self.verticalLayout_2.addLayout(self.horizontalLayout)
        self.preview_graphicsview = DocGraphicsView(self.widget1)
        self.preview_graphicsview.setObjectName("preview_graphicsview")
        self.verticalLayout_2.addWidget(self.preview_graphicsview)
        self.widget2 = QtWidgets.QWidget(self.splitter)
        self.widget2.setMinimumSize(QtCore.QSize(150, 0))
        self.widget2.setMaximumSize(QtCore.QSize(400, 16777215))
        self.widget2.setObjectName("widget2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.widget2)
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_3.setSpacing(0)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setSpacing(0)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.lineEdit_find = QtWidgets.QLineEdit(self.widget2)
        self.lineEdit_find.setObjectName("lineEdit_find")
        self.horizontalLayout_3.addWidget(self.lineEdit_find)
        self.verticalLayout_3.addLayout(self.horizontalLayout_3)
        self.thumb_graphicsview = ThumbGraphicsView(self.widget2)
        self.thumb_graphicsview.setMinimumSize(QtCore.QSize(0, 0))
        self.thumb_graphicsview.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.thumb_graphicsview.setObjectName("thumb_graphicsview")
        self.verticalLayout_3.addWidget(self.thumb_graphicsview)
        self.horizontalLayout_4.addWidget(self.splitter)

        self.retranslateUi(librarywidget)
        QtCore.QMetaObject.connectSlotsByName(librarywidget)

    def retranslateUi(self, librarywidget):
        _translate = QtCore.QCoreApplication.translate
        librarywidget.setWindowTitle(_translate("librarywidget", "Form"))
        self.tocButton_1.setText(_translate("librarywidget", " ⯈ "))
        self.label_of.setText(_translate("librarywidget", "  of"))
        self.label_pageCount.setText(_translate("librarywidget", "0 "))
from docgraphicsview import DocGraphicsView
from pagenolineedit import PageNoLineEdit
from thumbgraphicsview import ThumbGraphicsView
from tocpushbutton import TocPushButton
import resources_rc



from PyQt5 import QtCore
from PyQt5 import QtGui

# https://hzqtc.github.io/2012/04/poppler-vs-mupdf.html
# MuPDF is much faster at rendering, but slower at loading (get_page_sizes()) and also use more memories
# 
# TODO Mupdf freezes and crashes very often
BACKEND_MuPDF = False

if BACKEND_MuPDF:
    import fitz
    from popplerqt5 import Poppler # use poppler to load page sizes
else: 
    from popplerqt5 import Poppler

from multiprocessing import Process

from utils import debug
import time

class PdfRender(Process):

    # considering realtime, the request may be dropped
    # rendered = QtCore.pyqtSignal(str, int, float, QtGui.QImage)

    def __init__(self, commandQ, resultsQ):
        super(PdfRender, self).__init__()
        #
        self.commandQ = commandQ
        self.resultsQ = resultsQ

        self.doc = None
        self.filename = None
        # 
        self.painter = QtGui.QPainter()
        self.link_color = QtGui.QColor(0,0,127, 40)
        self.exit_flag = False

        self.requests_queue = {}
        
        # self.mutex = QtCore.QMutex()

    def set_document(self, filename):
        self.filename = filename
        self.requests_queue = {}
        
        if BACKEND_MuPDF:
            self.doc = fitz.open(self.filename)
            # use poppler to load page sizes (much faster)
            password = ""
            self.doc_poppler = Poppler.Document.load(self.filename, password.encode(), password.encode())
        else:
            password = ""
            self.doc = Poppler.Document.load(self.filename, password.encode(), password.encode())
            self.doc.setRenderHint(
                Poppler.Document.TextAntialiasing
                | Poppler.Document.TextHinting
                | Poppler.Document.Antialiasing
                )

    def get_page_sizes_mupdf(self, doc):
        pages_size_inch = []
        page_counts = len(doc)
        for i in range(page_counts):
            page_rect = doc[i].MediaBox
            pg_width = page_rect.width / 72.0 # width in inch
            pg_height = page_rect.height / 72.0
            pages_size_inch.append([pg_width, pg_height])
        return pages_size_inch

    def get_page_sizes_poppler(self, doc):
        pages_size_inch = []
        page_counts = doc.numPages()  
        for i in range(page_counts):
            pg_width = doc.page(i).pageSizeF().width() / 72.0 # width in inch
            pg_height = doc.page(i).pageSizeF().height() / 72.0
            pages_size_inch.append([pg_width, pg_height])  
        return pages_size_inch

    def get_page_sizes(self):
        # extract page sizes for all pages
        if BACKEND_MuPDF:
            # return self.get_page_sizes_mupdf(self.doc)
            return self.get_page_sizes_poppler(self.doc_poppler)
        else:
            return self.get_page_sizes_poppler(self.doc)

    def save_rendering_command(self, page_no, dpi, roi, visible_regions):
        if page_no in self.requests_queue:
            # remove old command with the same page number but different dpi                        
            dpi_list, roi_list = self.requests_queue[page_no]
            assert(len(dpi_list) == len(roi_list))
            cnt = len(dpi_list)
            duplicate_flag = False
            for i in range(cnt):
                if dpi_list[i] != dpi:
                    dpi_list[i] = None # mark deleted item as None
                    roi_list[i] = None
                elif roi_list[i] == roi:
                    duplicate_flag = True
            if not duplicate_flag:
                dpi_list.append(dpi)
                roi_list.append(roi)
        else:
            self.requests_queue[page_no] = [[dpi], [roi]]
            
        # only garentee current visible pages
        emptyKeys = []
        for page_no in self.requests_queue:
            if page_no not in visible_regions:
                emptyKeys.append(page_no)
                continue
            dpi_list, roi_list = self.requests_queue[page_no]
            assert(len(dpi_list) == len(roi_list))
            cnt = len(dpi_list)            
            for i in range(cnt):
                roi = roi_list[i]
                if roi is None:
                    continue
                if roi.intersected(visible_regions[page_no]).isEmpty():
                    roi_list[i] = None # mark deleted item as None
                    dpi_list[i] = None            
        for page_no in emptyKeys:
            self.requests_queue.pop(page_no)

    def get_command_from_queue(self):
        if len(self.requests_queue) == 0:
            return None
        emptyKeys = []
        command = None
        for page_no in self.requests_queue:
            dpi_list, roi_list = self.requests_queue[page_no]
            dpi = None
            roi = None
            while dpi is None:
                if len(dpi_list) == 0:
                    break
                dpi = dpi_list.pop(0)
                roi = roi_list.pop(0)
            if dpi is None:
                assert(len(dpi_list) == 0 and len(roi_list) == 0)
                emptyKeys.append(page_no)
            else:
                command = [page_no, dpi, roi]
                break
        for page_no in emptyKeys:
            self.requests_queue.pop(page_no)
        return command
        
    def get_toc_item_poppler(self, doc, node):
        element = node.toElement()
        title = element.tagName()
        # 
        linkDestination = None
        if element.hasAttribute("Destination"):
            linkDestination = Poppler.LinkDestination(element.attribute("Destination"))
        elif element.hasAttribute("DestinationName"):
            linkDestination = doc.linkDestination(element.attribute("DestinationName"))

        if linkDestination:
            # NOTE: in some files page_num may be in range 1 -> pages_count,
            # also, top may be not in range 0.0->1.0, we have to take care of that
            page_num = linkDestination.pageNumber()
            top = linkDestination.top() if linkDestination.isChangeTop() else 0
        else:
            page_num = -1
            top = -1
        return title, page_num, top

    def getTableOfContents(self):
        if BACKEND_MuPDF:
            return self.doc.getToC(simple=False)
        else:
            # Poppler API
            toc = self.doc.toc()
            if not toc:
                return []
            # construct toc from QDomDocument tree
            # each item will contain [level (the first entry is always 1), title, page (1-based), extra]
            # consistent with MuPDF
            toc_list = []
            current_level = 0
            nodes_queue = [[toc, current_level]]
            while len(nodes_queue) > 0:
                current_node, current_level = nodes_queue.pop(-1) # get the last one
                # 
                if current_level > 0:
                    # the root toc item contains nothing, the first valid level is always 1
                    title, page_num, _ = self.get_toc_item_poppler(self.doc, current_node)
                    toc_list.append([current_level, title, page_num, None])
                # 
                # push children in queue in reverse order
                current_node = current_node.lastChild()
                while not current_node.isNull():
                    nodes_queue.append([current_node, current_level + 1])
                    current_node = current_node.previousSibling()
            return toc_list

    def receive_commands(self):
        # collect all commands in queue
        size = self.commandQ.qsize()
        # debug("Qsize: ", size)
        for i in range(size):
             # will be blocked until when some data are avaliable
             # as we get qsize first, so there's no block here
            item = self.commandQ.get()
            command, params = item
            if command == 'SET':
                filename = params[0]
                self.set_document(filename)
            elif command == 'PAGESIZES':
                pages_size_inch = self.get_page_sizes()
                self.resultsQ.put(['PAGESIZES_RES', self.filename, pages_size_inch])
            elif command == 'TOC':
                toc = self.getTableOfContents()
                self.resultsQ.put(['TOC_RES', self.filename, toc])
            elif command == 'RENDER':
                page_no, dpi, roi, visible_regions =params
                self.save_rendering_command(page_no, dpi, roi, visible_regions)
            elif command == 'STOP':
                self.exit_flag = True
            else:
                # not supported command
                assert(0)

    def run(self):
        """ render(int, float)
        This slot takes page no. and dpi and renders that page, then emits a signal with QImage"""

        debug('render entered.')
        print(self.pid)

        while self.exit_flag == False:
            QtCore.QThread.msleep(20)

            self.receive_commands()

            # no file or no request even after reading pipe
            if self.doc is None or len(self.requests_queue) == 0:
                continue

            # render the first one in queue
            # debug("Request List: ", self.requests_list)
            
            command = self.get_command_from_queue()
            if command is None:
                continue
            
            page_no, dpi, roi = command
            
            if BACKEND_MuPDF:
                page = self.doc.loadPage(page_no)
            else:
                page = self.doc.page(page_no)

            if page is None:
                continue

            # debug('rendering page %d.' % (page_no))
            if BACKEND_MuPDF:
                page_rect = page.MediaBox
                # 
                zoom_ratio = dpi / 72.0
                x1 = roi.x() / zoom_ratio
                y1 = roi.y() / zoom_ratio
                x2 = (roi.x() + roi.width()) / zoom_ratio 
                y2 = (roi.y() + roi.height()) / zoom_ratio
                # 
                x1 = min(max(x1, 0), page_rect.width)
                y1 = min(max(y1, 0), page_rect.height)
                x2 = min(max(x2, 0), page_rect.width)
                y2 = min(max(y2, 0), page_rect.height)
                clip = fitz.Rect(x1, y1, x2, y2)
                pix = page.getPixmap(matrix=fitz.Matrix(zoom_ratio, zoom_ratio), clip=clip, alpha=False)
                # set the correct QImage format depending on alpha
                fmt = QtGui.QImage.Format_RGBA8888 if pix.alpha else QtGui.QImage.Format_RGB888
                img = QtGui.QImage(pix.samples, pix.width, pix.height, pix.stride, fmt)
                # 
                roi.setCoords(x1 * zoom_ratio, y1 * zoom_ratio, x2 * zoom_ratio, y2 * zoom_ratio) # write back roi
            else:
                pg_width_pix = dpi * (page.pageSizeF().width() / 72.0) # DPI * width in inch
                pg_height_pix = dpi * (page.pageSizeF().height() / 72.0)
                x1 = roi.x()
                y1 = roi.y()
                x2 = x1 + roi.width()
                y2 = y1 + roi.height()
                # 
                x1 = min(max(x1, 0), pg_width_pix)
                y1 = min(max(y1, 0), pg_height_pix)
                x2 = min(max(x2, 0), pg_width_pix)
                y2 = min(max(y2, 0), pg_height_pix)
                img = page.renderToImage(dpi, dpi, x1, y1, x2 - x1, y2 - y1)
                # 
                roi.setCoords(x1, y1, x2, y2) # write back roi
            

            # # Add Heighlight over Link Annotation
            # self.painter.begin(img)
            # annots = page.annotations()
            # for annot in annots:
            #     # if annot.subType() == Poppler.Annotation.ALink:
            #     x, y = annot.boundary().left()*img.width(), annot.boundary().top()*img.height()
            #     w, h = annot.boundary().width()*img.width()+1, annot.boundary().height()*img.height()+1
            #     self.painter.fillRect(x, y, w, h, self.link_color)
            # self.painter.end()
            # 
            # self.rendered.emit(self.filename, page_no, dpi, img)

            # QImage to QByteArray
            img_byte_array = QtCore.QByteArray()
            img_buffer = QtCore.QBuffer(img_byte_array)
            img_buffer.open(QtCore.QIODevice.WriteOnly)
            img.save(img_buffer, "png", quality=100)
            
            self.resultsQ.put(['RENDER_RES', self.filename, page_no, dpi, roi, img_byte_array])

        debug('render exited.')

class PdfReader(QtCore.QObject):
    """
    will be run in thread
    """
    # rendered = QtCore.pyqtSignal(int, QImage)
    # textFound = QtCore.pyqtSignal(int, list)
    annotationFound = QtCore.pyqtSignal(str, list)
    outlineFound = QtCore.pyqtSignal(QtGui.QStandardItemModel)

    def __init__(self):
        super(PdfReader, self).__init__()
        self.doc = None
        # self.page_set = page_set
        # self.painter = QPainter()
        # self.link_color = QColor(0,0,127, 40)

    def readAnnotation(self, filename):
        debug("start reading annotation")
        
        password = ""
        self.doc = Poppler.Document.load(filename, password.encode(), password.encode())
        self.doc.setRenderHint(
            Poppler.Document.TextAntialiasing
            | Poppler.Document.TextHinting
            | Poppler.Document.Antialiasing
            )

        page_num = self.doc.numPages()
        results = []
        for idx in range(page_num):
            
            # release the global interpreter lock (GIL) to make GUI interactable
            # if idx % 100 == 0:
            #     QtCore.QThread.msleep(10)

            pageAnnots = []
            doc_page = self.doc.page(idx)
            pg_width = doc_page.pageSize().width()
            pg_height = doc_page.pageSize().height()
            annots = doc_page.annotations()
            for item in annots:
                singleAnnot = {}
                singleAnnot['author'] = item.author()
                singleAnnot['boundary'] = item.boundary()
                singleAnnot['contents'] = item.contents()
                singleAnnot['modificationDate'] = item.modificationDate()
                singleAnnot['style'] = item.style()
                if item.subType() == Poppler.Annotation.AHighlight:
                    quads = item.highlightQuads()
                    txt = ""
                    for qd in quads:
                        x1 = qd.points[0].x() * pg_width
                        y1 = qd.points[0].y() * pg_height
                        x2 = qd.points[2].x() * pg_width
                        y2 = qd.points[2].y() * pg_height
                        # hacks: some pdfs have different meanings for these end points
                        rect = QtCore.QRectF(min(x1, x2), min(y1, y2), abs(x1 - x2), abs(y1 - y2))
                        txt = txt + doc_page.text(rect)
                        if len(txt) > 0 and txt[-1] == "-": # handle the last "-"
                            if txt[-2].islower():
                                txt = txt[:-1]
                        else:
                            txt += " "
                    singleAnnot['type'] = 'highlight'
                    singleAnnot['text'] = txt
                elif item.subType() == Poppler.Annotation.AGeom:
                    bounds = item.boundary()
                    custom_dpi = 150
                    page_pixel_width = custom_dpi * pg_width / 72.0
                    page_pixel_height = custom_dpi * pg_height / 72.0
                    roi = QtCore.QRectF(
                            bounds.left() * page_pixel_width, 
                            bounds.top() * page_pixel_height, 
                            bounds.width() * page_pixel_width, 
                            bounds.height() * page_pixel_height
                        )
                    
                    img = doc_page.renderToImage(custom_dpi, custom_dpi, roi.left(), roi.top(), roi.width(), roi.height())
                    singleAnnot['type'] = 'geom'
                    singleAnnot['image'] = img
                else:
                    continue # other types are not supported now
                pageAnnots.append(singleAnnot)
            results.append(pageAnnots)

        self.annotationFound.emit(filename, results)
        debug("finish reading annotation")

    def readOutline(self, doc, itemModel):
        parent_item = itemModel.invisibleRootItem()
        toc = doc.toc()
        if not toc:
            return
        node = toc.firstChild()
        loadOutline(doc, node, parent_item)
        self.outlineFound.emit(itemModel)

    def findText(self, doc, text, find_reverse):
        if find_reverse:
            pages = [i for i in range(1,page_num+1)]
            pages.reverse()
        else:
            pages = [i for i in range(page_num, self.doc.numPages()+1)]
        for page_no in pages:
            page = doc.page(page_no-1)
            textareas = page.search(text,Poppler.Page.CaseInsensitive,0)
            if textareas != []:
                self.textFound.emit(page_no, textareas)
                break

def loadOutline(doc, node, parent_item):
    """loadOutline(Poppler::Document* doc, const QDomNode& node, QStandardItem* parent_item) """
    element = node.toElement()
    item = QtGui.QStandardItem(element.tagName())

    linkDestination = None
    if element.hasAttribute("Destination"):
        linkDestination = Poppler.LinkDestination(element.attribute("Destination"))
    elif element.hasAttribute("DestinationName"):
        linkDestination = doc.linkDestination(element.attribute("DestinationName"))

    if linkDestination:
        # NOTE: in some files page_num may be in range 1 -> pages_count,
        # also, top may be not in range 0.0->1.0, we have to take care of that
        page_num = linkDestination.pageNumber()
        top = linkDestination.top() if linkDestination.isChangeTop() else 0

        item.setData(page_num, QtCore.Qt.UserRole + 1)
        item.setData(top, QtCore.Qt.UserRole + 2)

        pageItem = item.clone()
        pageItem.setText(str(page_num))
        pageItem.setTextAlignment(QtCore.Qt.AlignRight)

        parent_item.appendRow([item, pageItem])
    else:
        parent_item.appendRow(item)

    # Load next sibling
    siblingNode = node.nextSibling()
    if not siblingNode.isNull():
        loadOutline(doc, siblingNode, parent_item)

    # Load its child
    childNode = node.firstChild()
    if not childNode.isNull():
        loadOutline(doc, childNode, item)



from resources.ui_library import Ui_librarywidget
 
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from popplerqt5 import Poppler

from utils import debug
from pdfworker import PdfReader
from toc import TocManager

import os
import numpy as np
import json

class LibraryView(QtWidgets.QWidget, Ui_librarywidget):
    # renderRequested = QtCore.pyqtSignal(int, float)
    # loadFileRequested = QtCore.pyqtSignal(str, str)
    # findTextRequested = QtCore.pyqtSignal(str, int, bool)
    readAnnotationRequested = QtCore.pyqtSignal(str)
    # readOutlineRequested = QtCore.pyqtSignal(Poppler.Document, QtGui.QStandardItemModel)
    # findTextRequested = QtCore.pyqtSignal(str, int, bool)
    # pagePositionChanged = QtCore.pyqtSignal(int, int)
    showStatusRequested = QtCore.pyqtSignal(str)
    fileReselected = QtCore.pyqtSignal(str)

    def __init__(self, parent, screen_dpi, app_data_path):
        super(LibraryView, self).__init__(parent) # Call the inherited classes __init__ method
        self.setupUi(self)

        self.screen_dpi = screen_dpi
        self.app_data_path = app_data_path
        self.filename = ''

        # self.splitter_doc.setSizes([1, 0]) # set relative widths of its children
        # self.splitter_doc.setCollapsible(0, False) 
        self.pushButton_oneColumn.clicked.connect(self.onSetOneColumn)
        self.pushButton_twoColumn.clicked.connect(self.onSetTwoColumn)
        self.pushButton_fourColumn.clicked.connect(self.onSetFourColumn)
        self.pushButton_emptyPage.clicked.connect(self.onSetPrecedingEmptypage)
        self.pushButton_zoomIn.clicked.connect(self.zoomIn)
        self.pushButton_zoomOut.clicked.connect(self.zoomOut)
        self.pushButton_zoomFitWidth.clicked.connect(self.zoomFitWidth)

        # self.pushButton_oneColumn_2.clicked.connect(lambda:self.thumb_graphicsview.setColumnNumber(1))
        # self.pushButton_twoColumn_2.clicked.connect(lambda:self.thumb_graphicsview.setColumnNumber(2))
        # self.pushButton_fourColumn_2.clicked.connect(lambda:self.thumb_graphicsview.setColumnNumber(4))
        # self.pushButton_emptyPage_2.clicked.connect(self.onThumbSetPrecedingEmptypage)

        self.preview_graphicsview.viewColumnChanged.connect(self.onViewColumnChanged)
        # self.thumb_graphicsview.viewColumnChanged.connect(self.onThumbViewColumnChanged)

        self.preview_graphicsview.emptyLeadingPageChanged.connect(self.onEmptyLeadingPageChanged)
        # self.thumb_graphicsview.emptyLeadingPageChanged.connect(self.onThumbEmptyLeadingPageChanged)

        self.preview_graphicsview.zoomRatioChanged.connect(self.onZoomRatioChanged)

        self.preview_graphicsview.tocLoaded.connect(self.onTocLoaded)

        self.preview_graphicsview.viewportChanged.connect(self.onDocViewportChanged)
        # self.preview_graphicsview_2.viewportChanged.connect(self.onDocViewportChanged)

        self.lineEdit_pageNo.returnPressed.connect(self.onPageLineEditReturnPressed)

        self.preview_graphicsview.loadFinished.connect(self.onDoc1LoadFinished)
        self.thumb_graphicsview.loadFinished.connect(self.onThumbLoadFinished)
        self.thumb_graphicsview.pageRelocationRequest.connect(self.onThumbPageRelocationRequest)
        self.thumb_graphicsview.zoomRequest.connect(self.onThumbZoomRequest)

        # self.repos = ["/home/yhu/下载/", "/home/yhu/tmp/"]

        self.fileview_model = QtGui.QStandardItemModel(self)
        self.fileview.setModel(self.fileview_model)
        self.fileview.clicked.connect(self.onFileClick)

        self.fileview.setAlternatingRowColors(True)
        # self.annotview.setAlternatingRowColors(True)

        # self.annotview_model = QtGui.QStandardItemModel()
        # self.annotview.setModel(self.annotview_model)
        # self.annotationView.resizeRequested.connect(self.onAnnotaionViewResized)

        # # Create separate thread for reading pdf metainformations
        # self.pdf_reader_thread = QtCore.QThread(self)
        # self.pdf_reader_obj = PdfReader()
        # self.pdf_reader_obj.moveToThread(self.pdf_reader_thread) # this must be moved before connecting signals
        # # self.renderRequested.connect(self.pdf_reader_obj.render)
        # self.readAnnotationRequested.connect(self.pdf_reader_obj.readAnnotation)
        # # self.readOutlineRequested.connect(self.pdf_reader_obj.readOutline)
        # # self.findTextRequested.connect(self.pdf_reader_obj.findText)
        # self.pdf_reader_obj.annotationFound.connect(self.onAnnotationFound)
        # # self.pdf_reader_obj.outlineFound.connect(self.onOutlineFound)
        # # self.pdf_reader_obj.textFound.connect(self.onTextFound)
        # self.pdf_reader_thread.start()

        # dirname = self.repos[0]
        # self.setFileList(dirname)
        self.current_dir = None
        self.tocManager = TocManager(self.tocButton_1)
        self.tocManager.tocIndexChanged.connect(self.onTocIndexChanged)
        self.current_page_idx = -1
        self.pageCounts = 0

    def setFileList(self, dirname):
        # remove all file list first
        self.fileview_model.removeRows(0, self.fileview_model.rowCount())

        self.current_dir = dirname
        self.fileview_model.setHorizontalHeaderLabels([self.current_dir])

        parent_file_item = self.fileview_model.invisibleRootItem()

        pdf_filenames = os.listdir(dirname)
        for name in pdf_filenames:
            path = dirname + os.sep + name
            if os.path.isfile(path) and path.endswith('.pdf'):
                item = QtGui.QStandardItem(name)
                item.setData(name, QtCore.Qt.UserRole + 1)
                parent_file_item.appendRow(item)

    def setDocument(self, filename, screep_dpi):
        currentDir = os.path.dirname(filename)
        self.setFileList(currentDir)
        # 
        self.filename = filename
        self.screen_dpi = screep_dpi
        self.loadDocument(filename, screep_dpi)
        self.fileReselected.emit(self.filename)

    def onFileClick(self, m_index):
        # remove all annotatoin first
        # self.annotview_model.removeRows(0, self.annotview_model.rowCount())

        # save view status of the last file
        self.saveDocumentViewStatus(self.filename)

        self.filename = self.current_dir + os.sep + self.fileview_model.data(m_index, QtCore.Qt.UserRole + 1)
        print(self.filename)
        
        self.loadDocument(self.filename, self.screen_dpi)
        self.fileReselected.emit(self.filename)
                
    def loadDocument(self, filename, screen_dpi):
        viewStatus = self.loadDocumentViewStatus(filename)
        mainViewStatus = viewStatus['mainView'] if viewStatus else None
        self.preview_graphicsview.setDocument(filename, screen_dpi, mainViewStatus)
        thumbViewStatus = viewStatus['thumbView'] if viewStatus else None
        self.thumb_graphicsview.setDocument(filename, screen_dpi, thumbViewStatus)

    def onTocLoaded(self, toc):
        self.tocManager.setToc(toc)
        self.tocManager.update(self.current_page_idx)

    def onTocIndexChanged(self, page_no):
        debug("onTocIndexChanged: %d" % page_no)
        self.preview_graphicsview.gotoPage(page_no)

    def onDoc1LoadFinished(self):
        # debug("onDoc1LoadFinished")
        # self.thumb_graphicsview.setDocument(self.filename, self.screen_dpi)
        pass

    def onThumbLoadFinished(self):
        # debug("onThumbLoadFinished")
        pass

    def onDocViewportChanged(self, filename, page_counts, visible_regions):
        # debug("Doc viewport changed: ", visible_regions)
        self.thumb_graphicsview.highlightVisibleMasks(filename, visible_regions)
        # 
        self.current_page_idx = next(iter(visible_regions)) # fetch first key
        self.pageCounts = page_counts
        self.tocManager.update(self.current_page_idx)
        # 
        self.lineEdit_pageNo.setMaxLength(len(str(page_counts)))
        self.lineEdit_pageNo.setText("%d" % (self.current_page_idx + 1))
        self.label_pageCount.setText(" %d " % page_counts)

    def onPageLineEditReturnPressed(self):
        page_no = int(self.lineEdit_pageNo.text())
        self.gotoPage(page_no - 1)
        self.lineEdit_pageNo.clearFocus()

    def onThumbPageRelocationRequest(self, page_no, x_ratio, y_ratio):
        # debug("onPageRelocationRequest: <%d> (%.2f, %.2f)" % (page_no, x_ratio, y_ratio))
        viewRect = self.preview_graphicsview.viewport()
        self.preview_graphicsview.viewAtPageAnchor([page_no, x_ratio, y_ratio, viewRect.width()/2, viewRect.height()/2])

    def onThumbZoomRequest(self, zoomInFlag, page_no, x_ratio, y_ratio):
        # debug("onThumbZoomRequest: %d, <%d> (%.2f, %.2f)" % (zoomInFlag, page_no, x_ratio, y_ratio))
        viewRect = self.preview_graphicsview.viewport()
        if zoomInFlag:
            self.preview_graphicsview.zoomIn([page_no, x_ratio, y_ratio, viewRect.width()/2, viewRect.height()/2])
        else:
            self.preview_graphicsview.zoomOut([page_no, x_ratio, y_ratio, viewRect.width()/2, viewRect.height()/2])

    def onViewColumnChanged(self, viewColumn):
        if viewColumn == 1:
            self.pushButton_oneColumn.setChecked(True)
            self.pushButton_twoColumn.setChecked(False)
            self.pushButton_fourColumn.setChecked(False)
        elif viewColumn == 2:
            self.pushButton_oneColumn.setChecked(False)
            self.pushButton_twoColumn.setChecked(True)
            self.pushButton_fourColumn.setChecked(False)
        elif viewColumn == 4:
            self.pushButton_oneColumn.setChecked(False)
            self.pushButton_twoColumn.setChecked(False)
            self.pushButton_fourColumn.setChecked(True)
        else:
            assert(0)

    def onThumbViewColumnChanged(self, viewColumn):
        if viewColumn == 1:
            self.pushButton_oneColumn_2.setChecked(True)
            self.pushButton_twoColumn_2.setChecked(False)
            self.pushButton_fourColumn_2.setChecked(False)
        elif viewColumn == 2:
            self.pushButton_oneColumn_2.setChecked(False)
            self.pushButton_twoColumn_2.setChecked(True)
            self.pushButton_fourColumn_2.setChecked(False)
        elif viewColumn == 4:
            self.pushButton_oneColumn_2.setChecked(False)
            self.pushButton_twoColumn_2.setChecked(False)
            self.pushButton_fourColumn_2.setChecked(True)
        else:
            assert(0)
    
    def onZoomRatioChanged(self, zoomRatio):
        if zoomRatio == 0:
            self.pushButton_zoomFitWidth.setChecked(True)
            self.showStatusRequested.emit('Zoom fitted to width')
        else:
            self.pushButton_zoomFitWidth.setChecked(False)
            self.showStatusRequested.emit("Zoom to %d%%" % int(zoomRatio * 100))

    def onSetOneColumn(self):
        self.preview_graphicsview.setColumnNumber(1)
        self.thumb_graphicsview.setColumnNumber(1)

    def onSetTwoColumn(self):
        self.preview_graphicsview.setColumnNumber(2)
        self.thumb_graphicsview.setColumnNumber(2)

    def onSetFourColumn(self):
        self.preview_graphicsview.setColumnNumber(4)
        self.thumb_graphicsview.setColumnNumber(4)

    def onSetPrecedingEmptypage(self):
        if self.pushButton_emptyPage.isChecked():
            self.preview_graphicsview.setPrecedingEmptyPage(1)
            self.thumb_graphicsview.setPrecedingEmptyPage(1)
        else:
            self.preview_graphicsview.setPrecedingEmptyPage(0)
            self.thumb_graphicsview.setPrecedingEmptyPage(0) 

    def onEmptyLeadingPageChanged(self, emptyPages):
        if emptyPages == 0:
            self.pushButton_emptyPage.setChecked(False)
        elif emptyPages == 1:
            self.pushButton_emptyPage.setChecked(True)
        else:
            assert(0)

    def onAnnotationFound(self, filename, annotList):
        # the view has changed, too late
        debug(self.filename, filename)

        parent_item = self.annotview_model.invisibleRootItem()

        for pg_no in range(len(annotList)):
            for annot in annotList[pg_no]:
                image = None
                boundary = annot['boundary']
                contents = annot['contents']
                modificationDate = annot['modificationDate']
                item = QtGui.QStandardItem()
                item.setData(pg_no, QtCore.Qt.UserRole + 1)
                item.setData(boundary, QtCore.Qt.UserRole + 2)
                item.setData(contents, QtCore.Qt.UserRole + 3)
                item.setData(modificationDate, QtCore.Qt.UserRole + 4)
                # 
                timeStr = ("<%s>" % modificationDate.toString())
                if annot['type'] == 'highlight':
                    text = annot['text']
                    item.setText(text)
                elif annot['type'] == 'geom':
                    image = annot['image']
                    image = image.scaled(30, 30)
                    item.setData(QtGui.QPixmap.fromImage(image), QtCore.Qt.DecorationRole)
                    item.setData(image, QtCore.Qt.UserRole + 5)
                # item.setTextAlignment(QtCore.Qt.AlignLeft)

                dateItem = QtGui.QStandardItem()
                dateItem.setText(timeStr)
                dateItem.setTextAlignment(QtCore.Qt.AlignRight)

                parent_item.appendRow([item, dateItem])

    def goNextPage(self):
        self.preview_graphicsview.goNextPage()

    def goPrevPage(self):
        self.preview_graphicsview.goPrevPage()

    def gotoPage(self, pg_no):
        self.preview_graphicsview.gotoPage(pg_no)

    def zoomIn(self):
        self.preview_graphicsview.zoomIn()

    def zoomOut(self):
        self.preview_graphicsview.zoomOut()

    def zoomFitWidth(self):
        self.preview_graphicsview.zoomFitWidth()

    def closeEvent(self, ev):
        debug('closeEvent in LibraryView')
        self.saveDocumentViewStatus(self.filename)
        self.preview_graphicsview.close()
        self.thumb_graphicsview.close()

    def loadDocumentViewStatus(self, filename):
        dataFileName = self.app_data_path + os.sep + filename.replace(os.sep, '_') + '.json'
        try:
            with open(dataFileName, 'r') as f:
                status = json.load(f)
                return status
        except:
            return None

    def saveDocumentViewStatus(self, filename):
        # debug(self.preview_graphicsview.current_filename, filename)
        assert(self.preview_graphicsview.current_filename == filename)
        mainViewStatus = self.preview_graphicsview.getViewStatus()
        thumbViewStatus = self.thumb_graphicsview.getViewStatus()
        outDict = {
            "mainView": mainViewStatus,
            "thumbView": thumbViewStatus
        }
        dataFileName = self.app_data_path + os.sep + filename.replace(os.sep, '_') + '.json'
        with open(dataFileName, 'w') as f:
            json.dump(outDict, f, indent=True)


# -*- coding: utf-8 -*-

"""
Name = kuafu
Executable Command = kuafu
Package Name = kuafu
Python Module Name = kuafu
Debian Dependency = python3-pyqt5, python3-poppler-qt5

Description = A paper manger
Changes :
0.0.2   fixed :
0.0.1   added :
0.0.10   initial draft version

...........................................................................
|   Copyright (C) 2020-2021 Yinlin Hu <huyinlin@gmail.com>       |
|                                                                          |
|   This program is free software: you can redistribute it and/or modify   |
|   it under the terms of the GNU General Public License as published by   |
|   the Free Software Foundation, either version 3 of the License, or      |
|   (at your option) any later version.                                    |
|                                                                          |
|   This program is distributed in the hope that it will be useful,        |
|   but WITHOUT ANY WARRANTY; without even the implied warranty of         |
|   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          |
|   GNU General Public License for more details.                           |
|                                                                          |
|   You should have received a copy of the GNU General Public License      |
|   along with this program.  If not, see <http://www.gnu.org/licenses/>.  |
...........................................................................
"""

# TODO:
# There will be duplicated rendering when have multiple requests
# FIXME :

__version__ = '0.0.24'



from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui

from basedocgraphicsview import BaseDocGraphicsView
from utils import debug

class ThumbGraphicsView(BaseDocGraphicsView):
    pageRelocationRequest = QtCore.pyqtSignal(int, float, float)
    zoomRequest = QtCore.pyqtSignal(bool, int, float, float)

    def __init__(self, parent, render_num=4):
        super(ThumbGraphicsView, self).__init__(parent, render_num)

        self.scene.setBackgroundBrush(QtGui.QBrush(QtCore.Qt.white)) # set background

        self.setDragMode(QtWidgets.QGraphicsView.NoDrag) # disable the default dragger

    def highlightVisibleMasks(self, filename, visible_regions):
        # the items in visible_regions are in normalized coordinates
        # 
        if filename != self.current_filename:
            return
        if not self.load_finished_flag:
            return
        # 
        # remove current highlighted first
        for pg_no in self.current_highlighted_pages:
            self.page_items[pg_no].setMask(0, 0, 0, 0)
        self.current_highlighted_pages = []
        # remove border of the page that is marked as the current
        self.page_items[self.pageMarkedAsCurrent].setBorderHighlight(False)

        currentPageIdx = next(iter(visible_regions)) # first key as the current
        for pg_no in visible_regions:
            vRect = visible_regions[pg_no]
            self.initializePage(pg_no)
            if pg_no == currentPageIdx:
                self.page_items[pg_no].setBorderHighlight(True)
                self.pageMarkedAsCurrent = pg_no
            self.page_items[pg_no].setMask(vRect[0], vRect[1], vRect[2], vRect[3])
            if not self.isMouseOver:
                # self.centerOn(self.page_items[pg_no])
                self.ensureVisible(self.page_items[pg_no])
            # 
            self.current_highlighted_pages.append(pg_no)

    def mousePressEvent(self, ev):
        # debug('mousePressEvent in ThumbGraphicsView')
        if not self.load_finished_flag:
            return
        self.setCursor(QtCore.Qt.ClosedHandCursor)
        page_no, x_ratio, y_ratio, _, _ = self.getPageByPos(ev.pos().x(), ev.pos().y())
        self.pageRelocationRequest.emit(page_no, x_ratio, y_ratio)
        return super().mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        # debug('mouseReleaseEvent in ThumbGraphicsView')
        if not self.load_finished_flag:
            return
        self.setCursor(QtCore.Qt.ArrowCursor)
        return super().mouseReleaseEvent(ev)

    def mouseMoveEvent(self, ev):
        # debug('mouseMoveEvent in ThumbGraphicsView')
        if not self.load_finished_flag:
            return
        if self.isMousePressed:
            page_no, x_ratio, y_ratio, _, _ = self.getPageByPos(ev.pos().x(), ev.pos().y())
            self.pageRelocationRequest.emit(page_no, x_ratio, y_ratio)
        return super().mouseMoveEvent(ev)

    def wheelEvent(self, ev):
        # debug("wheelEvent in ThumbGraphicsView")
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        if modifiers == QtCore.Qt.ControlModifier:
            page_no, x_ratio, y_ratio, _, _ = self.getPageByPos(ev.pos().x(), ev.pos().y())
            delta = ev.angleDelta()
            if delta.y() > 0:
                self.zoomRequest.emit(True, page_no, x_ratio, y_ratio) # zoom in request
            else:
                self.zoomRequest.emit(False, page_no, x_ratio, y_ratio) # zoom out request
            ev.accept() # accept an event in order to stop it from propagating further
        else:
            return super().wheelEvent(ev) # call parent's handler, making the view scrolled (touch included)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file './kuafu/resources/ui_main.ui'
#
# Created by: PyQt5 UI code generator 5.15.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_window(object):
    def setupUi(self, window):
        window.setObjectName("window")
        window.resize(905, 722)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/kuafu.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        window.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(window)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.centraltabwidget = QtWidgets.QTabWidget(self.centralwidget)
        self.centraltabwidget.setTabBarAutoHide(True)
        self.centraltabwidget.setObjectName("centraltabwidget")
        self.verticalLayout_2.addWidget(self.centraltabwidget)
        window.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(window)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 905, 30))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuExport = QtWidgets.QMenu(self.menuFile)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/icons/export.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.menuExport.setIcon(icon1)
        self.menuExport.setObjectName("menuExport")
        self.menuRecentFiles = QtWidgets.QMenu(self.menuFile)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/icons/recents.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.menuRecentFiles.setIcon(icon2)
        self.menuRecentFiles.setObjectName("menuRecentFiles")
        self.menuNavigate = QtWidgets.QMenu(self.menubar)
        self.menuNavigate.setObjectName("menuNavigate")
        self.menuView = QtWidgets.QMenu(self.menubar)
        self.menuView.setObjectName("menuView")
        window.setMenuBar(self.menubar)
        self.statusBar = QtWidgets.QStatusBar(window)
        self.statusBar.setObjectName("statusBar")
        window.setStatusBar(self.statusBar)
        self.openFileAction = QtWidgets.QAction(window)
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/document-open.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.openFileAction.setIcon(icon3)
        self.openFileAction.setObjectName("openFileAction")
        self.quitAction = QtWidgets.QAction(window)
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/icons/im-ban-kick-user.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.quitAction.setIcon(icon4)
        self.quitAction.setObjectName("quitAction")
        self.zoominAction = QtWidgets.QAction(window)
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/icons/zoom-in.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoominAction.setIcon(icon5)
        self.zoominAction.setObjectName("zoominAction")
        self.zoomoutAction = QtWidgets.QAction(window)
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/icons/zoom-out.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomoutAction.setIcon(icon6)
        self.zoomoutAction.setObjectName("zoomoutAction")
        self.prevPageAction = QtWidgets.QAction(window)
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/icons/go-up.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.prevPageAction.setIcon(icon7)
        self.prevPageAction.setObjectName("prevPageAction")
        self.nextPageAction = QtWidgets.QAction(window)
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(":/icons/go-down.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.nextPageAction.setIcon(icon8)
        self.nextPageAction.setObjectName("nextPageAction")
        self.undoJumpAction = QtWidgets.QAction(window)
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap(":/icons/go-previous.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.undoJumpAction.setIcon(icon9)
        self.undoJumpAction.setObjectName("undoJumpAction")
        self.saveFileAction = QtWidgets.QAction(window)
        icon10 = QtGui.QIcon()
        icon10.addPixmap(QtGui.QPixmap(":/icons/document-save.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.saveFileAction.setIcon(icon10)
        self.saveFileAction.setObjectName("saveFileAction")
        self.toPSAction = QtWidgets.QAction(window)
        icon11 = QtGui.QIcon()
        icon11.addPixmap(QtGui.QPixmap(":/icons/adobe.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.toPSAction.setIcon(icon11)
        self.toPSAction.setObjectName("toPSAction")
        self.docInfoAction = QtWidgets.QAction(window)
        icon12 = QtGui.QIcon()
        icon12.addPixmap(QtGui.QPixmap(":/icons/info.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.docInfoAction.setIcon(icon12)
        self.docInfoAction.setObjectName("docInfoAction")
        self.pageToImageAction = QtWidgets.QAction(window)
        icon13 = QtGui.QIcon()
        icon13.addPixmap(QtGui.QPixmap(":/icons/image.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pageToImageAction.setIcon(icon13)
        self.pageToImageAction.setObjectName("pageToImageAction")
        self.printAction = QtWidgets.QAction(window)
        icon14 = QtGui.QIcon()
        icon14.addPixmap(QtGui.QPixmap(":/icons/document-print.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.printAction.setIcon(icon14)
        self.printAction.setObjectName("printAction")
        self.pageCountAction = QtWidgets.QAction(window)
        self.pageCountAction.setObjectName("pageCountAction")
        self.redoJumpAction = QtWidgets.QAction(window)
        icon15 = QtGui.QIcon()
        icon15.addPixmap(QtGui.QPixmap(":/icons/go-next.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.redoJumpAction.setIcon(icon15)
        self.redoJumpAction.setObjectName("redoJumpAction")
        self.zoomFitwidthAction = QtWidgets.QAction(window)
        self.zoomFitwidthAction.setCheckable(True)
        icon16 = QtGui.QIcon()
        icon16.addPixmap(QtGui.QPixmap(":/icons/zoom-fit-width.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomFitwidthAction.setIcon(icon16)
        self.zoomFitwidthAction.setObjectName("zoomFitwidthAction")
        self.undoModificationAction = QtWidgets.QAction(window)
        icon17 = QtGui.QIcon()
        icon17.addPixmap(QtGui.QPixmap(":/icons/edit-undo.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.undoModificationAction.setIcon(icon17)
        self.undoModificationAction.setObjectName("undoModificationAction")
        self.redoModificationAction = QtWidgets.QAction(window)
        icon18 = QtGui.QIcon()
        icon18.addPixmap(QtGui.QPixmap(":/icons/edit-redo.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.redoModificationAction.setIcon(icon18)
        self.redoModificationAction.setObjectName("redoModificationAction")
        self.actionFind = QtWidgets.QAction(window)
        icon19 = QtGui.QIcon()
        icon19.addPixmap(QtGui.QPixmap(":/icons/edit-find.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionFind.setIcon(icon19)
        self.actionFind.setObjectName("actionFind")
        self.actionViewInOneColumn = QtWidgets.QAction(window)
        self.actionViewInOneColumn.setCheckable(True)
        icon20 = QtGui.QIcon()
        icon20.addPixmap(QtGui.QPixmap(":/icons/snap-page.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionViewInOneColumn.setIcon(icon20)
        self.actionViewInOneColumn.setObjectName("actionViewInOneColumn")
        self.actionViewInTwoColumns = QtWidgets.QAction(window)
        self.actionViewInTwoColumns.setCheckable(True)
        icon21 = QtGui.QIcon()
        icon21.addPixmap(QtGui.QPixmap(":/icons/view-pages-facing.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionViewInTwoColumns.setIcon(icon21)
        self.actionViewInTwoColumns.setObjectName("actionViewInTwoColumns")
        self.actionPrecedingEmptyPage = QtWidgets.QAction(window)
        self.actionPrecedingEmptyPage.setCheckable(True)
        icon22 = QtGui.QIcon()
        icon22.addPixmap(QtGui.QPixmap(":/icons/page-simple.svg"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionPrecedingEmptyPage.setIcon(icon22)
        self.actionPrecedingEmptyPage.setObjectName("actionPrecedingEmptyPage")
        self.menuExport.addAction(self.pageToImageAction)
        self.menuExport.addAction(self.toPSAction)
        self.menuFile.addAction(self.openFileAction)
        self.menuFile.addAction(self.menuRecentFiles.menuAction())
        self.menuFile.addAction(self.menuExport.menuAction())
        self.menuFile.addAction(self.printAction)
        self.menuFile.addAction(self.docInfoAction)
        self.menuFile.addAction(self.quitAction)
        self.menuView.addAction(self.zoominAction)
        self.menuView.addAction(self.zoomoutAction)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuView.menuAction())
        self.menubar.addAction(self.menuNavigate.menuAction())

        self.retranslateUi(window)
        QtCore.QMetaObject.connectSlotsByName(window)

    def retranslateUi(self, window):
        _translate = QtCore.QCoreApplication.translate
        window.setWindowTitle(_translate("window", "kuafu"))
        self.menuFile.setTitle(_translate("window", "&File"))
        self.menuExport.setTitle(_translate("window", "Export"))
        self.menuRecentFiles.setTitle(_translate("window", "Recent Files"))
        self.menuNavigate.setTitle(_translate("window", "Help"))
        self.menuView.setTitle(_translate("window", "&View"))
        self.openFileAction.setText(_translate("window", "Open"))
        self.openFileAction.setShortcut(_translate("window", "Ctrl+O"))
        self.quitAction.setText(_translate("window", "Quit"))
        self.quitAction.setShortcut(_translate("window", "Ctrl+Q"))
        self.zoominAction.setText(_translate("window", "Zoom In"))
        self.zoominAction.setShortcut(_translate("window", "Ctrl++"))
        self.zoomoutAction.setText(_translate("window", "Zoom Out"))
        self.zoomoutAction.setShortcut(_translate("window", "Ctrl+-"))
        self.prevPageAction.setText(_translate("window", "Prev Page"))
        self.prevPageAction.setShortcut(_translate("window", "Left"))
        self.nextPageAction.setText(_translate("window", "Next Page"))
        self.nextPageAction.setShortcut(_translate("window", "Right"))
        self.undoJumpAction.setText(_translate("window", "Undo Jump"))
        self.undoJumpAction.setToolTip(_translate("window", "Undo Jump"))
        self.undoJumpAction.setShortcut(_translate("window", "Shift+Backspace"))
        self.saveFileAction.setText(_translate("window", "Save"))
        self.saveFileAction.setShortcut(_translate("window", "Ctrl+S"))
        self.toPSAction.setText(_translate("window", "to PostScript"))
        self.docInfoAction.setText(_translate("window", "Document Info"))
        self.pageToImageAction.setText(_translate("window", "Page to Image"))
        self.printAction.setText(_translate("window", "Print"))
        self.printAction.setToolTip(_translate("window", "Print with QuikPrint"))
        self.printAction.setShortcut(_translate("window", "Ctrl+P"))
        self.pageCountAction.setText(_translate("window", "0"))
        self.pageCountAction.setShortcut(_translate("window", "Ctrl+G"))
        self.redoJumpAction.setText(_translate("window", "Redo Jump"))
        self.redoJumpAction.setToolTip(_translate("window", "Redo Jump"))
        self.redoJumpAction.setShortcut(_translate("window", "Shift+Backspace"))
        self.zoomFitwidthAction.setText(_translate("window", "Zoom Fit Width"))
        self.zoomFitwidthAction.setToolTip(_translate("window", "Zoom Fit Width"))
        self.undoModificationAction.setText(_translate("window", "Undo Modification"))
        self.undoModificationAction.setToolTip(_translate("window", "Undo Modification"))
        self.undoModificationAction.setShortcut(_translate("window", "Ctrl+Z"))
        self.redoModificationAction.setText(_translate("window", "Redo Modification"))
        self.redoModificationAction.setToolTip(_translate("window", "Redo Modification"))
        self.redoModificationAction.setShortcut(_translate("window", "Ctrl+Y"))
        self.actionFind.setText(_translate("window", "Find"))
        self.actionFind.setToolTip(_translate("window", "Find"))
        self.actionFind.setShortcut(_translate("window", "Ctrl+F"))
        self.actionViewInOneColumn.setText(_translate("window", "View In One Column"))
        self.actionViewInOneColumn.setToolTip(_translate("window", "View In One Column"))
        self.actionViewInTwoColumns.setText(_translate("window", "View In Two Columns"))
        self.actionViewInTwoColumns.setToolTip(_translate("window", "View In Two Columns"))
        self.actionPrecedingEmptyPage.setText(_translate("window", "Preceding Empty Page"))
        self.actionPrecedingEmptyPage.setToolTip(_translate("window", "Preceding Empty Page"))
import resources_rc



 
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from resources.ui_annotation_item import Ui_annotationItemWidget

from popplerqt5 import Poppler

from utils import debug

class AnnotationFrame(QtWidgets.QFrame):
    """ This widget is a container of PageWidgets. PageWidget communicates
        Window through this widget """
    # jumpToRequested = QtCore.pyqtSignal(int, float)
    # copyTextRequested = QtCore.pyqtSignal(int, QtCore.QPoint, QtCore.QPoint)
    # showStatusRequested = QtCore.pyqtSignal(str)
    # pagePositionChanged = QtCore.pyqtSignal(int, int)

    # parent is scrollAreaWidgetContents
    def __init__(self, parent, scrollArea):
        super(AnnotationFrame, self).__init__(parent)

        self.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.setFrameShadow(QtWidgets.QFrame.Raised)

        self.vScrollbar = scrollArea.verticalScrollBar()
        self.hScrollbar = scrollArea.horizontalScrollBar()

        self.itemCount = 0

        self.setMouseTracking(True)
        self.clicked = False
        
        self.setContentsMargins(0, 0, 0, 0)

        # self.copy_text_mode = False

        # # 
        # self.doc = doc
        # self.pages_count = self.doc.numPages()
        # #
        # self.pre_empty_page_count = 0
        # self.view_column_count = 1

        self.layout = QtWidgets.QGridLayout(self)
        self.layout.setContentsMargins(0, 6, 0, 6) # left, top, right, bottom
        self.layout.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignCenter)
        # self.layout.setHorizontalSpacing(3)
        self.layout.setVerticalSpacing(15)

        # self.page_widgets = []
        # self.rearrangePages()

        # self.current_page = 0
        # self.current_render_dpi = 0 
        # self.zoom_fit_width = True # default on
        # self.screen_dpi = screen_dpi

        # self.visible_pages = [0]
        # self.rendered_pages = {}
        # self.scrollAreaSize = [0, 0] # width and height of the parent scroll area size

        # # create render threads
        # self.num_threads = threads
        # self.render_list = []
        # for i in range(self.num_threads):
        #     tmpRender = PdfRender()
        #     tmpRender.rendered.connect(self.setRenderedImage)
        #     tmpRender.start()
        #     self.render_list.append(tmpRender)

        # # set document in other threads
        # for rd in self.render_list:
        #     rd.set_document(self.doc)

    
    def addItem(self, color, author, date, title, comments):
        annotItem = AnnotationItemWidget(0, self)
        annotItem.setData(color, author, date, title, comments)
        self.layout.addWidget(annotItem, self.itemCount, 0, QtCore.Qt.AlignCenter)
        self.itemCount += 1

    def handleScrollAreaResized(self, width, height):
        debug('scrollAreaResized in AnnotationFrame')
        debug(width, height)
        width = width - 30
        itemCount = self.layout.count()
        for i in range(itemCount):
            self.layout.itemAt(i).widget().setFixedWidth(width)

    def showEvent(self, ev):
        debug('showEvent in AnnotationFrame')
        # QtWidgets.QWidget.showEvent(self, ev)
        # self.resizePages(self.current_render_dpi)

    def closeEvent(self, ev):
        """ Save all settings on window close """
        debug('closeEvent in AnnotationFrame')
        # for rd in self.render_list:
        #     rd.stop_async()
        # for rd in self.render_list:
        #     rd.wait()
        # return QtWidgets.QFrame.closeEvent(self, ev)

    def wheelEvent(self, ev):
        debug("wheelEvent in AnnotationFrame")
        # modifiers = QtWidgets.QApplication.keyboardModifiers()
        # if modifiers == QtCore.Qt.ControlModifier:
        #     delta = ev.angleDelta()
        #     if delta.y() > 0:
        #         self.zoomIn()
        #     else:
        #         self.zoomOut()
        #     ev.accept() # accept an event in order to stop it from propagating further

    def mousePressEvent(self, ev):
        debug("mousePressEvent in AnnotationFrame")
        self.click_pos = ev.globalPos()
        self.v_scrollbar_pos = self.vScrollbar.value()
        self.h_scrollbar_pos = self.hScrollbar.value()
        self.clicked = True

    def mouseReleaseEvent(self, ev):
        debug("mouseReleaseEvent in AnnotationFrame")
        self.clicked = False
        self.setCursor(QtCore.Qt.OpenHandCursor)

    def mouseMoveEvent(self, ev):
        # debug("mouseMoveEvent in DocFrame")
        if not self.clicked : return
        self.setCursor(QtCore.Qt.ClosedHandCursor)
        self.vScrollbar.setValue(self.v_scrollbar_pos + self.click_pos.y() - ev.globalY())
        self.hScrollbar.setValue(self.h_scrollbar_pos + self.click_pos.x() - ev.globalX())

    # def showStatus(self, msg):
    #     self.showStatusRequested.emit(msg)


class AnnotationItemWidget(QtWidgets.QWidget, Ui_annotationItemWidget):
    """ This widget shows a rendered page """
    def __init__(self, page_num, frame=None):
        super(AnnotationItemWidget, self).__init__(frame)
        self.setupUi(self)

        self.manager = frame
        self.setMouseTracking(True)
        self.setSizePolicy(0,0)
        self.link_areas = []
        self.link_annots = []
        self.annots_listed, self.copy_text_mode = False, False
        self.click_point, self.highlight_area = None, None
        self.page_num = page_num
        self.image = QtGui.QPixmap()

        self.groupBox.setContentsMargins(0, 0, 0, 0)

        # self.setStyleSheet("border-width: 1px; border-style: solid; border-color: black; background-color: white")
    
    def setData(self, color, author, date, title, comments):
        # 
        self.label_indicator.setAutoFillBackground(True)
        self.label_indicator.setText("")

        self.label_indicator.setStyleSheet(
            "background-color: rgb(%d,%d,%d)" % (color.red(), color.green(), color.blue()))

        self.groupBox.setStyleSheet("QGroupBox {border-width: 0px; border-style: solid}")

        self.label_author.setText(author)
        self.label_date.setText(date.toString())

        if isinstance(title, QtGui.QImage):
            self.label_title.setScaledContents(True)
            self.label_title.setPixmap(QtGui.QPixmap.fromImage(title))
        else:
            self.label_title.setText(title)
            # self.label_title.setMinimumSize(self.label_title.sizeHint())
            # self.label_title.setWordWrap(True)
        self.edit_comments.setText(comments)

    def clear(self):
        QtWidgets.QLabel.clear(self)
        self.setText("Loading ...")
        self.image = QtGui.QPixmap()

    # def resizeEvent(self, ev):
    #     debug("resizeEvent in AnnotationItemWidget")
    #     # print(self.width(), self.height())
    #     # QtWidgets.QScrollArea.resizeEvent(self, ev) # call parent's event handler
    #     # self.resizeRequested.emit(self.width(), self.height())

    # def mouseMoveEvent(self, ev):
    #     pass
    # def mousePressEvent(self, ev):
    #     pass

    # def mouseReleaseEvent(self, ev):
    #     pass

    def updateImage(self):
        """ repaint page widget, and draw highlight areas """
        if self.highlight_area:
            img = self.image.copy()
            painter = QPainter(img)
            zoom = self.dpi / 72.0
            for area in self.highlight_area:
                box = QtCore.QRectF(area.left()*zoom, area.top()*zoom,
                                    area.width()*zoom, area.height()*zoom)
                painter.fillRect(box, QColor(0,255,0, 127))
            painter.end()
            self.setPixmap(img.scaled(self.width(), self.height()))
            # self.setPixmap(img)
        else:
            # self.setPixmap(self.image)
            self.setPixmap(self.image.scaled(self.width(), self.height()))




from setuptools import setup
from kuafu import __version__

setup(
      name='kuafu',
      version=__version__,
      description='PDF manager',
      keywords='PDF',
      url='',
      author='Yinlin Hu',
      author_email='huyinlin@gmail.com',
      license='GNU GPLv3',
      packages=['kuafu'],
      install_requires=['PyQt5', 'numpy', 'opencv-python', 'pymupdf'], #external packages as dependencies
      classifiers=[
      'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
      'Operating System :: POSIX :: Linux',
      'Programming Language :: Python :: 3',
      ],
      entry_points={
          'console_scripts': ['kuafu=kuafu.main:main'],
      },
      data_files=[
                 ('share/applications', ['kuafu/resources/kuafu.desktop']),
                 ('share/icons', ['kuafu/resources/icons/kuafu.svg']),
      ],
      include_package_data=True,
      zip_safe=False
)
